@@ -7,7 +7,7 @@ import zigpy_znp.types as t
 from zigpy_znp.types import nvids
 
 
-class SysCommands(t.CommandsBase, subsystem=t.Subsystem.SYS):
+class Sys(t.CommandsBase, subsystem=t.Subsystem.SYS):
     # reset the target device
     ResetReq = t.CommandDef(
         t.CommandType.AREQ,
@@ -16,7 +16,7 @@ class SysCommands(t.CommandsBase, subsystem=t.Subsystem.SYS):
             (
                 t.Param(
                     "Type",
-                    t.uint8_t,
+                    t.ResetType,
                     (
                         "This command will reset the device by using a hardware reset "
                         "(i.e. watchdog reset) if ‘Type’ is zero. Otherwise a soft "
