@@ -12,14 +12,16 @@ import zigpy.application
 import zigpy.profiles
 import zigpy.zcl.foundation
 
-from zigpy.zdo.types import ZDOCmd
+from zigpy.zdo.types import ZDOCmd, ZDOHeader, CLUSTERS as ZDO_CLUSTERS
 
-from zigpy.types import ExtendedPanId
+from zigpy.types import ExtendedPanId, deserialize as list_deserialize
 from zigpy.zcl.clusters.security import IasZone
+from zigpy.exceptions import DeliveryError
 
 import zigpy_znp.config as conf
 import zigpy_znp.types as t
 import zigpy_znp.commands as c
+from zigpy_znp.exceptions import InvalidCommandResponse
 
 from zigpy_znp.api import ZNP
 from zigpy_znp.types.nvids import NwkNvIds
@@ -33,46 +35,45 @@ LOGGER = logging.getLogger(__name__)
 
 ZDO_CONVERTERS = {
     ZDOCmd.Node_Desc_req: (
-        ZDOCmd.Node_Desc_rsp,
         (
-            lambda addr, ep: c.ZDOCommands.NodeDescReq.Req(
-                DstAddr=addr, NWKAddrOfInterest=addr
+            lambda addr, device, NWKAddrOfInterest: c.ZDO.NodeDescReq.Req(
+                DstAddr=addr, NWKAddrOfInterest=NWKAddrOfInterest
             )
         ),
-        (
-            lambda addr: c.ZDOCommands.NodeDescRsp.Callback(
-                partial=True, Src=addr, Status=t.ZDOStatus.SUCCESS
-            )
-        ),
-        (lambda rsp, dev: [rsp.NodeDescriptor]),
+        (lambda addr: c.ZDO.NodeDescRsp.Callback(partial=True, Src=addr)),
+        (lambda rsp: (ZDOCmd.Node_Desc_rsp, [rsp.NodeDescriptor])),
     ),
     ZDOCmd.Active_EP_req: (
-        ZDOCmd.Active_EP_rsp,
         (
-            lambda addr, ep: c.ZDOCommands.ActiveEpReq.Req(
-                DstAddr=addr, NWKAddrOfInterest=addr
+            lambda addr, device, NWKAddrOfInterest: c.ZDO.ActiveEpReq.Req(
+                DstAddr=addr, NWKAddrOfInterest=NWKAddrOfInterest
             )
         ),
-        (
-            lambda addr: c.ZDOCommands.ActiveEpRsp.Callback(
-                partial=True, Src=addr, Status=t.ZDOStatus.SUCCESS
-            )
-        ),
-        (lambda rsp, dev: [rsp.ActiveEndpoints]),
+        (lambda addr: c.ZDO.ActiveEpRsp.Callback(partial=True, Src=addr)),
+        (lambda rsp: (ZDOCmd.Active_EP_rsp, [rsp.ActiveEndpoints])),
     ),
     ZDOCmd.Simple_Desc_req: (
-        ZDOCmd.Simple_Desc_rsp,
         (
-            lambda addr, ep: c.ZDOCommands.SimpleDescReq.Req(
-                DstAddr=addr, NWKAddrOfInterest=addr, Endpoint=ep
+            # fmt: off
+            lambda addr, device, NWKAddrOfInterest, EndPoint: \
+            c.ZDO.SimpleDescReq.Req(
+                DstAddr=addr, NWKAddrOfInterest=NWKAddrOfInterest, Endpoint=EndPoint
+            )
+            # fmt: on
+        ),
+        (lambda addr: c.ZDO.SimpleDescRsp.Callback(partial=True, Src=addr)),
+        (lambda rsp: (ZDOCmd.Simple_Desc_rsp, [rsp.SimpleDescriptor])),
+    ),
+    ZDOCmd.Mgmt_Leave_req: (
+        (
+            lambda addr, device, DeviceAddress, Options: c.ZDO.MgmtLeaveReq.Req(
+                DstAddr=addr,
+                IEEE=device.ieee,
+                RemoveChildren_Rejoin=c.zdo.LeaveOptions(Options),
             )
         ),
-        (
-            lambda addr: c.ZDOCommands.SimpleDescRsp.Callback(
-                partial=True, Src=addr, Status=t.ZDOStatus.SUCCESS
-            )
-        ),
-        (lambda rsp, dev: [rsp.SimpleDescriptor]),
+        (lambda addr: c.ZDO.MgmtLeaveRsp.Callback(partial=True, Src=addr)),
+        (lambda rsp: (ZDOCmd.Mgmt_Leave_rsp, [rsp.Status])),
     ),
 }
 
@@ -106,26 +107,24 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         finally:
             znp.close()
 
-    def on_zdo_relays_message(self, msg: c.ZDOCommands.SrcRtgInd.Callback) -> None:
+    def on_zdo_relays_message(self, msg: c.ZDO.SrcRtgInd.Callback) -> None:
         LOGGER.info("ZDO device relays: %s", msg)
         device = self.get_device(nwk=msg.DstAddr)
         device.relays = msg.Relays
 
-    def on_zdo_device_announce(
-        self, msg: c.ZDOCommands.EndDeviceAnnceInd.Callback
-    ) -> None:
+    def on_zdo_device_announce(self, msg: c.ZDO.EndDeviceAnnceInd.Callback) -> None:
         LOGGER.info("ZDO device announce: %s", msg)
         self.handle_join(nwk=msg.NWK, ieee=msg.IEEE, parent_nwk=0x0000)
 
-    def on_zdo_device_join(self, msg: c.ZDOCommands.TCDevInd.Callback) -> None:
+    def on_zdo_device_join(self, msg: c.ZDO.TCDevInd.Callback) -> None:
         LOGGER.info("ZDO device join: %s", msg)
         self.handle_join(nwk=msg.SrcNwk, ieee=msg.SrcIEEE, parent_nwk=msg.ParentNwk)
 
-    def on_zdo_device_leave(self, msg: c.ZDOCommands.LeaveInd.Callback) -> None:
+    def on_zdo_device_leave(self, msg: c.ZDO.LeaveInd.Callback) -> None:
         LOGGER.info("ZDO device left: %s", msg)
         self.handle_leave(nwk=msg.NWK, ieee=msg.IEEE)
 
-    def on_af_message(self, msg: c.AFCommands.IncomingMsg.Callback) -> None:
+    def on_af_message(self, msg: c.AF.IncomingMsg.Callback) -> None:
         try:
             device = self.get_device(nwk=msg.SrcAddr)
         except KeyError:
@@ -153,26 +152,24 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
     def _bind_callbacks(self, api):
         api.callback_for_response(
-            c.AFCommands.IncomingMsg.Callback(partial=True), self.on_af_message
+            c.AF.IncomingMsg.Callback(partial=True), self.on_af_message
         )
 
         # ZDO requests need to be handled explicitly
         api.callback_for_response(
-            c.ZDOCommands.EndDeviceAnnceInd.Callback(partial=True),
-            self.on_zdo_device_announce,
+            c.ZDO.EndDeviceAnnceInd.Callback(partial=True), self.on_zdo_device_announce,
         )
 
         api.callback_for_response(
-            c.ZDOCommands.TCDevInd.Callback.Callback(partial=True),
-            self.on_zdo_device_join,
+            c.ZDO.TCDevInd.Callback.Callback(partial=True), self.on_zdo_device_join,
         )
 
         api.callback_for_response(
-            c.ZDOCommands.LeaveInd.Callback(partial=True), self.on_zdo_device_leave
+            c.ZDO.LeaveInd.Callback(partial=True), self.on_zdo_device_leave
         )
 
         api.callback_for_response(
-            c.ZDOCommands.SrcRtgInd.Callback(partial=True), self.on_zdo_relays_message
+            c.ZDO.SrcRtgInd.Callback(partial=True), self.on_zdo_relays_message
         )
 
     async def _reconnect(self) -> None:
@@ -220,7 +217,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         output_clusters=[],
     ):
         return await self._znp.request(
-            c.AFCommands.Register.Req(
+            c.AF.Register.Req(
                 Endpoint=endpoint,
                 ProfileId=profile_id,
                 DeviceId=device_id,
@@ -256,35 +253,54 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         await self._znp.nvram_write(NwkNvIds.SRC_RTG_EXPIRY_TIME, t.uint8_t(255))
         await self._znp.nvram_write(NwkNvIds.NWK_CHILD_AGE_ENABLE, t.Bool(False))
 
-        await self._reset()
+        try:
+            is_configured = (
+                await self._znp.nvram_read(NwkNvIds.HAS_CONFIGURED_ZSTACK3)
+            ) == b"\x55"
+        except InvalidCommandResponse as e:
+            assert e.response.Status == t.Status.InvalidParameter
+            is_configured = False
 
-        if auto_form and False:
-            # XXX: actually form a network
+        if not is_configured and not auto_form:
+            raise RuntimeError("Cannot start application, network is not formed")
+        elif auto_form and is_configured:
+            LOGGER.info("ZNP is already configured, no need to form a network.")
+        elif auto_form and not is_configured:
             await self.form_network()
 
         if self.config[conf.CONF_ZNP_CONFIG][conf.CONF_TX_POWER] is not None:
             dbm = self.config[conf.CONF_ZNP_CONFIG][conf.CONF_TX_POWER]
 
             await self._znp.request(
-                c.SysCommands.SetTxPower.Req(TXPower=dbm), RspStatus=t.Status.Success
+                c.Sys.SetTxPower.Req(TXPower=dbm), RspStatus=t.Status.Success
+            )
+
+        device_info = await self._znp.request(
+            c.Util.GetDeviceInfo.Req(), RspStatus=t.Status.Success
+        )
+
+        if device_info.DeviceState != t.DeviceState.StartedAsCoordinator:
+            # Start the application and wait until it's ready
+            await self._znp.request_callback_rsp(
+                request=c.ZDO.StartupFromApp.Req(StartDelay=100),
+                RspState=c.zdo.StartupState.RestoredNetworkState,
+                callback=c.ZDO.StateChangeInd.Callback(
+                    State=t.DeviceState.StartedAsCoordinator
+                ),
             )
 
-        """
         # Get our active endpoints
         endpoints = await self._znp.request_callback_rsp(
-            request=c.ZDOCommands.ActiveEpReq.Req(
-                DstAddr=0x0000, NWKAddrOfInterest=0x0000
-            ),
+            request=c.ZDO.ActiveEpReq.Req(DstAddr=0x0000, NWKAddrOfInterest=0x0000),
             RspStatus=t.Status.Success,
-            callback=c.ZDOCommands.ActiveEpRsp.Callback(partial=True),
+            callback=c.ZDO.ActiveEpRsp.Callback(partial=True),
         )
 
         # Clear out the list of active endpoints
         for endpoint in endpoints.ActiveEndpoints:
             await self._znp.request(
-                c.AFCommands.Delete(Endpoint=endpoint), RspStatus=t.Status.Success
+                c.AF.Delete.Req(Endpoint=endpoint), RspStatus=t.Status.Success
             )
-        """
 
         # Register our endpoints
         await self._register_endpoint(endpoint=1)
@@ -299,17 +315,6 @@ class ControllerApplication(zigpy.application.ControllerApplication):
             endpoint=100, profile_id=zigpy.profiles.zll.PROFILE_ID, device_id=0x0005
         )
 
-        # Start commissioning and wait until it's done
-        await self._znp.request_callback_rsp(
-            request=c.APPConfigCommands.BDBStartCommissioning.Req(
-                Mode=c.app_config.BDBCommissioningMode.NetworkFormation
-            ),
-            RspStatus=t.Status.Success,
-            callback=c.APPConfigCommands.BDBCommissioningNotification.Callback(
-                partial=True, Status=c.app_config.BDBCommissioningStatus.Success,
-            ),
-        )
-
     async def update_network(
         self,
         *,
@@ -341,15 +346,14 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
         if channels is not None:
             await self._znp.request(
-                c.UtilCommands.SetChannels.Req(Channels=channels),
+                c.Util.SetChannels.Req(Channels=channels), RspStatus=t.Status.Success,
+            )
+            await self._znp.request(
+                c.AppConfig.BDBSetChannel.Req(IsPrimary=True, Channel=channels),
                 RspStatus=t.Status.Success,
             )
             await self._znp.request(
-                c.APPConfigCommands.BDBSetChannel.Req(IsPrimary=True, Channel=channels),
-                RspStatus=t.Status.Success,
-            )
-            await self._znp.request(
-                c.APPConfigCommands.BDBSetChannel.Req(
+                c.AppConfig.BDBSetChannel.Req(
                     IsPrimary=False, Channel=t.Channels.NO_CHANNELS
                 ),
                 RspStatus=t.Status.Success,
@@ -359,7 +363,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
         if pan_id is not None:
             await self._znp.request(
-                c.UtilCommands.SetPanId.Req(PanId=pan_id), RspStatus=t.Status.Success
+                c.Util.SetPanId.Req(PanId=pan_id), RspStatus=t.Status.Success
             )
 
             self._pan_id = pan_id
@@ -372,7 +376,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
         if network_key is not None:
             await self._znp.request(
-                c.UtilCommands.SetPreConfigKey.Req(PreConfigKey=network_key),
+                c.Util.SetPreConfigKey.Req(PreConfigKey=network_key),
                 RspStatus=t.Status.Success,
             )
 
@@ -385,8 +389,8 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
     async def _reset(self):
         await self._znp.request_callback_rsp(
-            request=c.SysCommands.ResetReq.Req(Type=t.ResetType.Soft),
-            callback=c.SysCommands.ResetInd.Callback(partial=True),
+            request=c.Sys.ResetReq.Req(Type=t.ResetType.Soft),
+            callback=c.Sys.ResetInd.Callback(partial=True),
         )
 
     async def form_network(self):
@@ -399,36 +403,29 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         await self._znp.nvram_write(
             NwkNvIds.LOGICAL_TYPE, t.DeviceLogicalType.Coordinator
         )
-        await self._reset()
 
-        # If zgPreConfigKeys is set to TRUE, all devices should use the same
-        # pre-configured security key. If zgPreConfigKeys is set to FALSE, the
-        # pre-configured key is set only on the coordinator device, and is handed to
-        # joining devices. The key is sent in the clear over the last hop. Upon reset,
-        # the device will retrieve the pre-configured key from NV memory if the NV_INIT
-        # compile option is defined (the NV item is called ZCD_NV_PRECFGKEY).
-
-        pan_id = self.config[conf.SCHEMA_NETWORK][conf.CONF_NWK_PAN_ID]
-        extended_pan_id = self.config[conf.SCHEMA_NETWORK][
-            conf.CONF_NWK_EXTENDED_PAN_ID
-        ]
+        pan_id = self.config[conf.CONF_NWK][conf.CONF_NWK_PAN_ID]
+        extended_pan_id = self.config[conf.CONF_NWK][conf.CONF_NWK_EXTENDED_PAN_ID]
 
         await self.update_network(
-            channels=self.config[conf.SCHEMA_NETWORK][conf.CONF_NWK_CHANNELS],
+            channels=self.config[conf.CONF_NWK][conf.CONF_NWK_CHANNELS],
             pan_id=0xFFFF if pan_id is None else pan_id,
             extended_pan_id=ExtendedPanId(os.urandom(8))
             if extended_pan_id is None
             else extended_pan_id,
-            network_key=t.KeyData([os.urandom(16)]),
+            network_key=t.KeyData(os.urandom(16)),
             reset=False,
         )
 
         # We want to receive all ZDO callbacks to proxy them back go zipgy
         await self._znp.nvram_write(NwkNvIds.ZDO_DIRECT_CB, t.Bool(True))
 
+        # Reset now so that the changes take effect
+        await self._reset()
+
         await self._znp.request(
-            c.APPConfigCommands.BDBStartCommissioning.Req(
-                Mode=c.app_config.BDBCommissioningMode.NetworkFormation
+            c.AppConfig.BDBStartCommissioning.Req(
+                Mode=c.app_config.BDBCommissioningMode.NwkFormation
             ),
             RspStatus=t.Status.Success,
         )
@@ -436,18 +433,31 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         # This may take a while because of some sort of background scanning.
         # This can probably be disabled.
         await self._znp.wait_for_response(
-            c.ZDOCommands.StateChangeInd.Callback(
-                State=t.DeviceState.StartedAsCoordinator
-            )
+            c.ZDO.StateChangeInd.Callback(State=t.DeviceState.StartedAsCoordinator)
         )
 
         await self._znp.request(
-            c.APPConfigCommands.BDBStartCommissioning.Req(
-                Mode=c.app_config.BDBCommissioningMode.NetworkSteering
+            c.AppConfig.BDBStartCommissioning.Req(
+                Mode=c.app_config.BDBCommissioningMode.NwkSteering
             ),
             RspStatus=t.Status.Success,
         )
 
+        # Create the NV item that keeps track of whether or not we're configured
+        osal_create_rsp = await self._znp.request(
+            c.Sys.OSALNVItemInit.Req(
+                Id=NwkNvIds.HAS_CONFIGURED_ZSTACK3, ItemLen=1, Value=b""
+            )
+        )
+
+        if osal_create_rsp.Status not in (t.Status.Success, t.Status.ItemCreated):
+            raise RuntimeError(
+                "Could not create HAS_CONFIGURED_ZSTACK3 NV item"
+            )  # pragma: no cover
+
+        # Initializing the item won't guarantee that it holds this exact value
+        await self._znp.nvram_write(NwkNvIds.HAS_CONFIGURED_ZSTACK3, b"\x55")
+
     async def _send_zdo_request(
         self, dst_addr, dst_ep, src_ep, cluster, sequence, options, radius, data
     ):
@@ -474,25 +484,36 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
         assert dst_ep == ZDO_ENDPOINT
 
-        rsp_cluster, req_factory, callback_factory, converter = ZDO_CONVERTERS[cluster]
-        request = req_factory(dst_addr.address, ep=src_ep)
-        callback = callback_factory(dst_addr.address)
-
-        LOGGER.debug(
-            "Intercepted AP ZDO request and replaced with %s - %s", request, callback
-        )
-
-        async with async_timeout.timeout(ZDO_REQUEST_TIMEOUT):
-            response = await self._znp.request_callback_rsp(
-                request=request, RspStatus=t.Status.Success, callback=callback
-            )
+        # Deserialize the ZDO request
+        zdo_hdr, data = ZDOHeader.deserialize(cluster, data)
+        field_names, field_types = ZDO_CLUSTERS[cluster]
+        zdo_args, _ = list_deserialize(data, field_types)
+        zdo_kwargs = dict(zip(field_names, zdo_args))
 
         device = self.get_device(nwk=dst_addr.address)
 
+        # Call the converter with the ZDO request's kwargs
+        req_factory, rsp_factory, zdo_rsp_factory = ZDO_CONVERTERS[cluster]
+        request = req_factory(dst_addr.address, device, **zdo_kwargs)
+        callback = rsp_factory(dst_addr.address)
+
+        LOGGER.debug(
+            "Intercepted AP ZDO request and replaced with %s/%s", request, callback
+        )
+
+        try:
+            async with async_timeout.timeout(ZDO_REQUEST_TIMEOUT):
+                response = await self._znp.request_callback_rsp(
+                    request=request, RspStatus=t.Status.Success, callback=callback
+                )
+        except InvalidCommandResponse as e:
+            raise DeliveryError(f"Could not send command: {e.response.Status}") from e
+
+        zdo_rsp_cluster, zdo_response_args = zdo_rsp_factory(response)
+
         # Build up a ZDO response
         message = t.serialize_list(
-            [t.uint8_t(sequence), response.Status, response.NWK]
-            + converter(response, device)
+            [t.uint8_t(sequence), response.Status, response.NWK] + zdo_response_args
         )
         LOGGER.trace("Pretending we received a ZDO message: %s", message)
 
@@ -500,7 +521,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         self.handle_message(
             sender=device,
             profile=zigpy.profiles.zha.PROFILE_ID,
-            cluster=rsp_cluster,
+            cluster=zdo_rsp_cluster,
             src_ep=dst_ep,
             dst_ep=src_ep,
             message=message,
@@ -519,26 +540,34 @@ class ControllerApplication(zigpy.application.ControllerApplication):
                 dst_addr, dst_ep, src_ep, cluster, sequence, options, radius, data
             )
 
-        async with async_timeout.timeout(DATA_CONFIRM_TIMEOUT):
-            response = await self._znp.request_callback_rsp(
-                request=c.AFCommands.DataRequestExt.Req(
-                    DstAddrModeAddress=dst_addr,
-                    DstEndpoint=dst_ep,
-                    DstPanId=0x0000,
-                    SrcEndpoint=src_ep,
-                    ClusterId=cluster,
-                    TSN=sequence,
-                    Options=options,
-                    Radius=radius,
-                    Data=data,
-                ),
-                RspStatus=t.Status.Success,
-                callback=c.AFCommands.DataConfirm.Callback(
-                    partial=True, Endpoint=dst_ep, TSN=sequence
-                ),
-            )
+        request = c.AF.DataRequestExt.Req(
+            DstAddrModeAddress=dst_addr,
+            DstEndpoint=dst_ep,
+            DstPanId=0x0000,
+            SrcEndpoint=src_ep,
+            ClusterId=cluster,
+            TSN=sequence,
+            Options=options,
+            Radius=radius,
+            Data=data,
+        )
 
-        LOGGER.debug("Received a data request confirmation: %s", response)
+        if dst_addr.mode == t.AddrMode.Broadcast:
+            # We won't always get a data confirmation
+            response = await self._znp.request(
+                request=request, RspStatus=t.Status.Success
+            )
+        else:
+            async with async_timeout.timeout(DATA_CONFIRM_TIMEOUT):
+                response = await self._znp.request_callback_rsp(
+                    request=request,
+                    RspStatus=t.Status.Success,
+                    callback=c.AF.DataConfirm.Callback(
+                        partial=True, Endpoint=dst_ep, TSN=sequence
+                    ),
+                )
+
+            LOGGER.debug("Received a data request confirmation: %s", response)
 
         if response.Status != t.Status.Success:
             return response.Status, "Invalid response status"
@@ -623,24 +652,31 @@ class ControllerApplication(zigpy.application.ControllerApplication):
     async def force_remove(self, device) -> None:
         """Forcibly remove device from NCP."""
         await self._znp.request(
-            c.ZDOCommands.MgmtLeaveReq.Req(
-                DstAddr=device.nwk,
+            c.ZDO.MgmtLeaveReq.Req(
+                DstAddr=0x0000,  # We handle it
                 IEEE=device.ieee,
-                LeaveOptions=c.zdo.LeaveOptions.NONE,
+                RemoveChildren_Rejoin=c.zdo.LeaveOptions.NONE,
             ),
             RspStatus=t.Status.Success,
         )
 
+        # TODO: see what happens when we forcibly remove a device that isn't our child
+
+        # Just wait for the response, removing the device will be handled upstream
+        await self._znp.wait_for_response(
+            c.ZDO.LeaveInd.Callback(NWK=device.nwk, IEEE=device.ieee, partial=True)
+        )
+
     async def permit_ncp(self, time_s: int) -> None:
         response = await self._znp.request_callback_rsp(
-            request=c.ZDOCommands.MgmtPermitJoinReq.Req(
+            request=c.ZDO.MgmtPermitJoinReq.Req(
                 AddrMode=t.AddrMode.Broadcast,
                 Dst=zigpy.types.BroadcastAddress.ALL_DEVICES,
                 Duration=time_s,
                 TCSignificance=0,  # not used in Z-Stack
             ),
             RspStatus=t.Status.Success,
-            callback=c.ZDOCommands.MgmtPermitJoinRsp.Callback(partial=True),
+            callback=c.ZDO.MgmtPermitJoinRsp.Callback(partial=True),
         )
 
         if response.Status != t.Status.Success:
